module S4U
	module S4u_Linetool
		class LineTool
			def initialize(line_style)
				@ip1 = nil
				@ip2 = nil
				ext=IS_OSX ? ".pdf" : ".svg"
				cursor_path= File.join(File.dirname(__FILE__),("icon_mouse_pen" + ext))
				@cursor_id = UI.create_cursor(cursor_path,0,30) if cursor_path
				cursor_path1= File.join(File.dirname(__FILE__),("icon_mouse" + ext))
				@cursor_id1 = UI.create_cursor(cursor_path1,0,0) if cursor_path1
				@strings=S4U::S4u_Linetool::Strings
				@snap=20
				@line_style=0
				@line_style=line_style if line_style.between?(0,6)
				@lock_z=S4U::S4u_Linetool.instance_variable_get(:@lock_z)
				@scale_factor= Sketchup.version.to_i > 17 ? UI.scale_factor : 1
				@select_element=false
			end
			def activate
				model=Sketchup.active_model
				@edit_transform= axes_edit_transform
				model.active_view.lock_inference
				@ip1 = Sketchup::InputPoint.new
				@ip2 = Sketchup::InputPoint.new
				@ip = Sketchup::InputPoint.new
				@drawn = false
				Sketchup::set_status_text @strings.GetString("Length"), SB_VCB_LABEL
				@components=model.selection.grep(Sketchup::ComponentInstance)
				@components=model.selection.grep(Sketchup::Group) if @components.empty?
				@select_element=false
				if @components.empty? && @line_style==6
					# UI.messagebox Strings["Select 1 Component or Group"]
					@select_element=true
				end
				self.reset(nil)
				@last_z=nil
			end
			def axes_edit_transform
				model=Sketchup.active_model
				return model.axes.transformation if Sketchup.version.to_i>=16
				model.start_operation "get edit transform"
				ents=model.active_entities
				begin
					g=ents.add_group(ents.add_group)
					t=Geom::Transformation.new(g.transformation.xaxis,g.transformation.yaxis,g.transformation.zaxis,g.transformation.origin)
					g.erase! if g && g.valid? && !g.deleted?
				rescue
					t=model.edit_transform
				end
				model.abort_operation
				return t
			end
			def onSetCursor
				if @line_style==6 && (@select_element || @components.empty? || @alt_press)
					UI.set_cursor(@cursor_id1) if @cursor_id1
				else
					UI.set_cursor(@cursor_id) if @cursor_id
				end
			end
			def getExtents
				bb = Sketchup.active_model.bounds
				bb.add @ip1.position if @ip1.valid?
				bb.add @ip2.position if @ip2.valid?
				return bb
			end
			def getMenu(menu)
				menus=[]
				menus<< menu.add_item(@strings.GetString("Draw Lines")){ @line_style=0 ;onSetCursor }
				menus<< menu.add_item(@strings.GetString("Draw Construction Lines: ..........")){ @line_style=1 ;onSetCursor }
				menus<< menu.add_item(@strings.GetString("Draw Construction Lines: ------")){ @line_style=2 ;onSetCursor }
				menus<< menu.add_item(@strings.GetString("Draw Construction Lines: _ _ _ _")){ @line_style=3 ;onSetCursor }
				menus<< menu.add_item(@strings.GetString("Draw Construction Lines: -.- -.-")){ @line_style=4  ;onSetCursor}
				menu.add_separator
				menus<< menu.add_item(@strings.GetString("Draw Construction Point")){ self.reset;@line_style=5 ;onSetCursor }
				menus<< menu.add_item(@strings.GetString("Place Component")){
					self.reset;
					@line_style=6
					if @components.empty?
						@select_element=true
						# UI.messagebox Strings["Select 1 Component or Group"]
					end
				}
				menus.each{|m| menu.set_validation_proc(m) { MF_UNCHECKED }}
				menu.set_validation_proc(menus[@line_style]) { MF_CHECKED }
				menu.add_separator
				menus<< menu.add_item(@strings.GetString("Lock Z")){ @lock_z=1-@lock_z }
				menu.set_validation_proc(menus[-1]) { @lock_z==1 ? MF_CHECKED : MF_UNCHECKED }
			end
			def deactivate(view)
				S4U::S4u_Linetool.instance_variable_set(:@line_style,@line_style) if @line_style.between?(0,6)
				S4U::S4u_Linetool.instance_variable_set(:@lock_z,@lock_z)
				view.invalidate if @drawn
			end
			def get_vertices(face)
				return nil if not( face && (face.is_a?Sketchup::Face))
				ver=[]
				vertices=[]
				vertices =face.outer_loop.vertices*3
				(0..face.outer_loop.vertices.length-1).each { |i|
					j=face.outer_loop.vertices.length+i
					p1=vertices[j-1].position.clone
					p2=vertices[j].position.clone
					p3=vertices[j+1].position.clone
					v1=p1.vector_to p2
					v2=p2.vector_to p3
					ver.push vertices[j] if not( v1.samedirection? v2)
				}
				return ver
			end
			def set_pointonedge(x,y,view,ip,edge)
				p1=edge.start.position.clone
				p2=edge.end.position.clone
				points= [Geom.linear_combination(2.0/3, p1,1.0/3, p2),Geom.linear_combination(1.0/3, p1,2.0/3, p2)]
				ph = view.pick_helper
				ph.init(x, y)
				picked = points.find{|p| ph.test_point((p.transform ip.transformation),x,y,@snap*@scale_factor) }
				if picked
					ip1=Sketchup::InputPoint.new(picked.transform ip.transformation)
					ip.copy! ip1
					view.tooltip="       " + @strings.GetString("1/3")
					return ip
				end
				return nil
			end
			def center_face(face,num=200)
				return nil if face.outer_loop.vertices.length> num #|| face.loops.length>1
				vertices=get_vertices(face)
				return nil if !vertices
				numver=vertices.length
				cp=Geom::Point3d.new
				vertices.each{|v|
					cp.x=cp.x+v.position.x
					cp.y=cp.y+v.position.y
					cp.z=cp.z+v.position.z
				}
				cp.x=cp.x/ numver
				cp.y=cp.y/ numver
				cp.z=cp.z/ numver
				return cp
			end
			def set_pointonedge(x,y,view,ip,edge)
				p1=edge.start.position.clone
				p2=edge.end.position.clone
				return nil if edge.length < (view.pixels_to_model(5*@snap*@scale_factor, p1))
				points= [Geom.linear_combination(2.0/3, p1,1.0/3, p2),Geom.linear_combination(1.0/3, p1,2.0/3, p2)]
				ph = view.pick_helper
				ph.init(x, y)
				picked = points.find{|p| ph.test_point((p.transform ip.transformation),x,y,@snap*@scale_factor) }
				if picked
					ip1=Sketchup::InputPoint.new(picked.transform ip.transformation)
					ip.copy! ip1
					view.tooltip= @strings.GetString("1/3")
					return ip
				end
				return nil
			end
			def get_bounds(element)
				b=nil
				if element.is_a?Sketchup::ComponentInstance
					b=element.definition.bounds
				elsif element.is_a?Sketchup::Group
					b=element.local_bounds
				elsif element.is_a?Sketchup::Image
					b=element.bounds
				else
					return nil
				end
				return (0..7).collect{|i| b.corner(i).clone.transform element.transformation } if b
			end
			def set_special_point(x,y,view,ip)
				ph = view.pick_helper
				ph.init(x, y)
				ph.do_pick(x, y)
				sels=view.model.selection.select{|e| e.valid? && (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance) || e.is_a?(Sketchup::Image))}
				points=[]
				@picked_element << ph.best_picked  if (e=ph.best_picked) && (e.valid? && (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance)))
				sels+=@picked_element
				best_entity = sels.sort_by{|e| view.screen_coords(e.bounds.center).distance view.screen_coords(ip.position) }
				best_entity=best_entity[0..7].select{|e| e.valid? && (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance) )}
				if !best_entity.empty?
					best_entity.each{|e|
						ps = get_bounds(e)
						points+=ps
						ps.permutation(2).to_a.each{|a|  points << Geom.linear_combination(0.5, a[0],0.5,a[1]) }
						# get_lines_bounds(e).each{|a|  points << Geom.linear_combination(0.5, a[0],0.5,a[1]) }
					}
				end
				points.uniq!

				picked_bounds = points.find_all{|p| ph.test_point(p,x,y,1.5*@snap*@scale_factor) }
				if picked_bounds
					picked_bounds.each{|picked|
						model=view.model
						ray=view.pickray x, y
						if ray
							ray=[picked,ray[1].reverse!]
							item=model.raytest(ray,false)
							if !item
								@edge=true
								ip1=Sketchup::InputPoint.new(picked)
								ip.copy! ip1
								view.tooltip=@strings.GetString("BoundingBox")
								return picked
							end
						end
					}
				end
				if ip && ip.valid? && ip.face
					cf=center_face(ip.face)
					return ip if !cf
					cf.transform! ip.transformation
					ph = view.pick_helper
					ph.init(x, y)
					picked = ph.test_point(cf,x,y,@snap*@scale_factor)
					if picked
						ip1=Sketchup::InputPoint.new(cf)
						@face=true
						view.tooltip=@strings.GetString("Center face")
						ip.copy! ip1
					else
						@edge=ip.face.edges.find{|e| set_pointonedge(x,y,view,ip,e) }
					end
				elsif ip.edge
					@edge=set_pointonedge(x,y,view,ip,ip.edge)
				elsif @center.length>0
					@center.each{|cf|
						cf.transform! ip.transformation
						ph = view.pick_helper
						ph.init(x, y)
						picked = ph.test_point(cf,x,y,@snap*5*@scale_factor)
						if picked
							ip1=Sketchup::InputPoint.new(cf)
							@center_point=true
							view.tooltip=@strings.GetString("Center")
							ip.copy! ip1
						end
					}
				end
				return ip
			end
			def onMouseMove(flags, x, y, view)
				@shift_press= (CONSTRAIN_MODIFIER_MASK&flags==CONSTRAIN_MODIFIER_MASK)
				@ctrl_press= (COPY_MODIFIER_MASK&flags==COPY_MODIFIER_MASK)
				@alt_press= (ALT_MODIFIER_MASK&flags==ALT_MODIFIER_MASK)
				@face=@edge=nil
				@center_point=nil
				if @select_element && @components.empty? && @line_style==6
					Sketchup::set_status_text Strings["Select 1 Component or Group"],SB_PROMPT
				else
					if( @state == 0 )
						@ip.pick view, x, y
						if @ip && @ip.valid? && @ip.edge && @ip.edge.curve
							begin
								curve=@ip.edge.curve
								c,vector,radius,a,vertices,reverse=get_info(curve)
								@center<< c if c
								@center.uniq!{|p| [p.x,p.y,p.z] }
							rescue=>error
								p error
							end
						end
						if @ip != @ip1
							view.tooltip = @ip.tooltip
							set_special_point(x,y,view,@ip) #if !view.inference_locked?
							view.invalidate
							@ip1.copy! @ip
							ip=@ip.position.transform @edit_transform.inverse
							line=@ip.position-@edit_transform.origin
							text=[ip.x.to_s , ip.y.to_s , ip.z.to_s]
							pos=[@ip.position.x.to_s , @ip.position.y.to_s , @ip.position.z.to_s].join(",")
							localpos=[ip.x.to_s , ip.y.to_s , ip.z.to_s].join(",")
							if Sketchup.version.to_i>=14
								@text= @strings.GetString("1st") + " | " +
								@strings.GetString("Global") + " : " + pos + " | " +
								@strings.GetString("Local") + " : " + localpos + " | " +

								@strings.GetString("Global Angle") + " : " +
								(@ip.position-ORIGIN).angle_between(X_AXIS).radians.round(1).to_s + " , " +
								(@ip.position-ORIGIN).angle_between(Y_AXIS).radians.round(1).to_s + " , " +
								(0.5*Math::PI-(@ip.position-ORIGIN).angle_between(Z_AXIS)).radians.round(1).to_s  + " | " +

								@strings.GetString("Local Angle") + " : " +
								line.angle_between(@edit_transform.xaxis).radians.round(1).to_s + " , " +
								line.angle_between(@edit_transform.yaxis).radians.round(1).to_s + " , " +
								(0.5*Math::PI-line.angle_between(@edit_transform.zaxis)).radians.round(1).to_s
								Sketchup::set_status_text @text, SB_PROMPT
							end
						end
					else
						Sketchup::set_status_text @strings.GetString("Length"), SB_VCB_LABEL
						ip=@ip.position.transform @edit_transform.inverse
						line=@ip.position-@ip1.position
						pos=[@ip.position.x.to_s , @ip.position.y.to_s , @ip.position.z.to_s].join(",")
						localpos=[ip.x.to_s , ip.y.to_s , ip.z.to_s].join(",")
						if Sketchup.version.to_i>=14
							t="";t1=""
							if @ip.position.z != @ip1.position.z
								ratio=Math::tan(0.5*Math::PI-(@ip.position-@ip1.position).angle_between(Z_AXIS))
								t= ratio<=1 ? ( " (" + ratio.round(1).to_s + "/1.0)" ) : (" (" +  ratio.round(1).to_s  + "/1.0)")
							end
							if @ip.position.z != @ip1.position.z
								ratio=Math::tan(0.5*Math::PI-line.angle_between(@edit_transform.zaxis))
								t1= ratio<=1 ? ( " (" + ratio.round(1).to_s + "/1.0)" ) : (" (" +  ratio.round(1).to_s  + "/1.0)")
							end
							@text= @strings.GetString("2nd") + " | " +
							@strings.GetString("Global") + " : " + pos + " | " +
							@strings.GetString("Local") + " : " + localpos + " | " +

								@strings.GetString("Global Angle") + " : " +
								(@ip.position-@ip1.position).angle_between(X_AXIS).radians.round(1).to_s + " , " +
								(@ip.position-@ip1.position).angle_between(Y_AXIS).radians.round(1).to_s + " , " +
								(0.5*Math::PI-(@ip.position-@ip1.position).angle_between(Z_AXIS)).radians.round(1).to_s +
								t.to_s + " | " +

								@strings.GetString("Local Angle") + " : " +
								line.angle_between(@edit_transform.xaxis).radians.round(1).to_s + " , " +
								line.angle_between(@edit_transform.yaxis).radians.round(1).to_s + " , " +
								(0.5*Math::PI-line.angle_between(@edit_transform.zaxis)).radians.round(1).to_s +
								t1.to_s
							Sketchup::set_status_text @text, SB_PROMPT
						end
						@ip.pick view, x, y, @ip1
						view.tooltip = @ip.tooltip if( @ip.valid? )
						set_special_point(x,y,view,@ip) #if !view.inference_locked?
						if @shift_down_time && @p_shift && @ip!=@ip1 && @ip2!=@ip1
							p2_line=@ip.position.project_to_line([@ip1.position,@p_shift])
							@ip2.copy!(p2_line ? Sketchup::InputPoint.new(p2_line) : @ip)
						else
							@ip2.copy! @ip
						end
						view.invalidate
						if( @ip2.valid? )
							length = @ip1.position.distance(@ip2.position)
							Sketchup::set_status_text length.to_s, SB_VCB_VALUE
						end
					end
				end
			end
			def onLButtonDown(flags, x, y, view)
				if ((@select_element && @components.empty? ) || @alt_press) && @line_style==6
					ph = view.pick_helper
					ph.do_pick x,y
					entity = ph.best_picked
					if entity &&  entity.is_a?(Sketchup::ComponentInstance) || entity.is_a?(Sketchup::Group)
						@components.clear
						@components<<entity
						@select_element=false
					else
						UI.messagebox Strings["Select 1 Component or Group"]
					end
				else
					if( @state == 0 )
						if( @ip1.valid? )
							if @line_style==6
								self.create_geometry(@ip1.position, @ip1.position, view)
								self.reset(view)
							else
								if @line_style==5
									p = @ip1.position
									if @alt_press
										if @center.length>0
											p=@center[0]
											p p
											self.create_geometry(p,p, view)
											self.reset(view)
										else
											p "no center"
										end
									else
										self.create_geometry(p,p, view)
										self.reset(view)
									end
								else
									if @alt_press
										if @center.length>0
											p=@center[0]
											@ip1.copy! Sketchup::InputPoint.new(p)
											@state = 1
											Sketchup::set_status_text @strings.GetString("Select second end"), SB_PROMPT
										end
									else
										@state = 1
										Sketchup::set_status_text @strings.GetString("Select second end"), SB_PROMPT
									end
								end
							end
						end
					else
						if( @ip2.valid? )
							if @alt_press
								if @center.length>0
									p=@center[0]
									@ip2.copy! Sketchup::InputPoint.new(p)
									self.create_geometry(@ip1.position, @ip2.position,view)
									@ip1.copy! @ip2
								end
							else
								ip = Sketchup::InputPoint.new @ip2.position
								self.create_geometry(@ip1.position, @ip2.position,view)
								if ip.valid?
									@ip1.copy! ip
								else
									self.reset(view)
								end
							end
						end
					end
					view.lock_inference
				end
			end
			def onKeyDown(key, repeat, flags, view)
				if( key == CONSTRAIN_MODIFIER_KEY )
					@shift_down_time =true
					if( view.inference_locked? )
						view.lock_inference
					elsif( @state == 0 && @ip1.valid? )
						view.lock_inference @ip1
					elsif( @state == 1 && @ip2.valid? )
						@p_shift=@ip.position
						view.lock_inference @ip2, @ip1
					end
				elsif key==VK_UP
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p=  @ip1.position
						p1=p + @edit_transform.zaxis
						ip1 = Sketchup::InputPoint.new(p1)
						view.lock_inference ip1,@ip
						@ip_lock=ip1.position
						@ip2.copy! ip1
					end
				elsif key==VK_LEFT
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p1=@ip1.position + @edit_transform.yaxis
						ip1 = Sketchup::InputPoint.new(p1)
						view.lock_inference ip1,@ip
						@ip_lock=ip1.position
						@ip2.copy! ip1
					end
				elsif key==VK_RIGHT
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p1=@ip1.position + @edit_transform.xaxis
						ip1 = Sketchup::InputPoint.new(p1)
						view.lock_inference ip1,@ip
						@ip_lock=ip1.position
						@ip2.copy! ip1
					end
				elsif (key==VK_ALT && IS_WIN )|| ((key==VK_CONTROL) && IS_OSX)
					@alt_press=true
				end
			end
			def onKeyUp(key, repeat, flags, view)
				if( key == CONSTRAIN_MODIFIER_KEY )
					#&& view.inference_locked? &&
					#(Time.now - @shift_down_time) > 0.5 )
					@shift_down_time=nil
					@p_shift=nil
					view.lock_inference
				elsif ((key==VK_CONTROL) && IS_WIN ) || ((key==VK_ALT) && IS_OSX)
					@ctrl_press=nil
				elsif (key==VK_ALT && IS_WIN )|| ((key==VK_CONTROL) && IS_OSX)
					@alt_press=nil
				end
			end
			def onUserText(text, view)
				if @state==0
					v= text.split ','
					p=[0,0,0]
					if v.length==2
						x,y= cal(v[0]),cal(v[1])
						(p.x,p.y = x,y) if x && y
					elsif v.length==3
						x,y,z= cal(v[0]),cal(v[1]),cal(v[2])
						(p.x,p.y,p.z = x,y,z) if x && y && z
					end
					@ip1=Sketchup::InputPoint.new(p)
					@state=1
				elsif @state == 1
					return if not @ip2.valid?
					pt1 = @ip1.position
					vec = @ip2.position - pt1
					v= text.split ','
					p=[0,0,0]
					if v.length==2
						x,y= cal(v[0]),cal(v[1])
						(p.x,p.y = x,y) if x && y
						pt2=Geom::Point3d.new p
					elsif v.length==3
						x,y,z= cal(v[0]),cal(v[1]),cal(v[2])
						(p.x,p.y,p.z = x,y,z) if x && y && z
						pt2=Geom::Point3d.new p
					else
						value=cal(text)
						return if !value
						if( vec.length != 0.0 )
							vec.length = value
							pt2 = pt1 + vec
						end
					end
					Sketchup::set_status_text value, SB_VCB_VALUE
					self.create_geometry(pt1, pt2, view)
					@ip1.copy! Sketchup::InputPoint.new(pt2)
					#self.reset(view)
				end
			end
			def draw(view)
				if @ip && @ip.valid?
					if @ip.display?
						@ip.draw(view)
					elsif @face || @edge
						view.draw_points @ip.position, 9*@scale_factor, 1, "red"
					elsif @center_point
						view.draw_points @ip.position, 9*@scale_factor, 1, "red"
					end
				end
				if( @ip1.valid? )
					if( @ip1.display? )
						@ip1.draw(view)
						@drawn = true
					end
					if @ip2.valid? && @line_style<5
						@drawn = true
						if @p_shift ||((@face || @edge) && @ip_lock)
							p=@p_shift
							p = @ip_lock if (@face || @edge) && @ip_lock
							p2_line=@ip.position.project_to_line([@ip1.position,p])
							view.draw_points p2_line, 9*@scale_factor, 2, "red"  if p2_line
							@ip2.copy! Sketchup::InputPoint.new(p2_line)
							view.line_stipple = "."
							view.drawing_color ="blue"
							p1,p2=@ip1.position, @ip2.position
							if @lock_z==1
								p1.z=@last_z ? @last_z : p1.z
								p2.z=p1.z
							end
							view.draw_line(p1,p2)
							view.line_stipple = ""
							view.drawing_color ="black"
						end
						@ip2.draw(view) if( @ip2.display? )
						view.set_color_from_line(@ip1, @ip2)
						line_style=["",".","-","_","-.-"]
						view.line_stipple = line_style[@line_style]
						p1,p2=@ip1.position, @ip2.position
						if @lock_z==1
							p1.z=@last_z ? @last_z : p1.z
							p2.z=p1.z
						end
						view.draw_line(p1,p2)
					end
				end
			end
			def onCancel(flag, view)
				self.reset(view)
			end
			def reset(view=nil)
				@state = 0
				@picked_element=[]
				Sketchup::set_status_text(@strings.GetString("Select first end"), SB_PROMPT)
				@ip1.clear
				@ip2.clear
				@center=[]
				@ip_lock=nil

				if @components.empty? && @line_style==6
					@select_element=true
				else
					@select_element=false
				end
				if( view )
					view.tooltip = nil
					view.invalidate if @drawn
				end
				@drawn = false
			end
			def cal(input)
				lengthformat=Sketchup.active_model.options["UnitsOptions"]["LengthFormat"]
				if op=input[/[\+\-\/\*]/]
					begin
					if input[/[\+\-]/]
						index= input.rindex(/[\+\-]/)
					elsif input[/[\*\/]/]
						index= input.rindex(/[\*\/]/)
					end
					if index
						if index>0
							s=[input[0..(index-1)],input[(index+1)..-1]]
						else
							s=["0",input[(index+1)..-1]]
						end
						op=input[index]
						if (lengthformat==0 || lengthformat==2)
							if op=="+"
								value= cal(s[0]).to_s.to_f + cal(s[1]).to_s.to_f
							elsif op=="-"
								value= cal(s[0]).to_s.to_f - cal(s[1]).to_s.to_f
							elsif op=="*"
								value= cal(s[0]).to_s.to_f * cal(s[1]).to_s.to_f
							elsif op=="/"
								value= cal(s[0]).to_s.to_f / cal(s[1]).to_s.to_f
							end
						else
							if op=="+"
								value= cal(s[0]).to_f + cal(s[1]).to_f
							elsif op=="-"
								value= cal(s[0]).to_f - cal(s[1]).to_f
							elsif op=="*"
								value= cal(s[0]).to_f * cal(s[1]).to_f
							elsif op=="/"
								value= cal(s[0]).to_f / cal(s[1]).to_f
							end
						end
					end
					rescue
						value=parse_length(input)
					end
				else
					value = parse_length(input)
				end
				return parse_length(value)
			end
			def parse_length(input)

				(return 0) if !input
				input=input.to_s
				if !Sketchup.parse_length("1.1")
					input.gsub!(/\./, ',')
				else
					input.gsub!(/\,/, '.')
				end
				begin
					value=input.to_l
				rescue
					value=input.to_f
				end
				return value
			end
			def create_geometry(p1, p2, view)
				model=Sketchup.active_model
				model.start_operation(@strings.GetString("Draw line"),true,false,false)
				faces=model.active_entities.grep(Sketchup::Face)
				edges=nil
				if  @line_style==0
					g=model.active_entities.add_group
					if @lock_z==1 || @ctrl_press
						p1.z=@last_z ? @last_z : p1.z
						p2.z=p1.z
					end
					g.entities.add_line(p1,p2)
					@last_z=p2.z
					edges= g.explode.grep(Sketchup::Edge)
				elsif @line_style==5
					model.active_entities.add_cpoint p1
				elsif @line_style==6
					if @components.empty?
						# @select_element=true
						UI.messagebox Strings["Select 1 Component or Group"]
						model.abort_operation
					else
						if @ctrl_press
							p1.z= @last_z ? @last_z : p1.z
						end
						trans=Geom::Transformation.new p1
						model.active_entities.add_instance @components[0].definition,trans
						@last_z=p1.z
					end
				elsif @line_style<5
					line_style=["",".","-","_","-.-"]
					edge= model.active_entities.add_cline p1,p2,line_style[@line_style]
					model.active_entities.add_cpoint p1 if !@cp2 || (@cp2 && @cp2.valid? && @cp2.position!=p1)
					@cp2=model.active_entities.add_cpoint p2
				end
				if edges && edges.length>0 && @line_style==0
					edges.each{|e| e.find_faces if e && e.valid? }
					pend=(edges.collect{|e| e.vertices}.flatten).min_by{|m| m.position.distance p2}
					faces=model.active_entities.grep(Sketchup::Face)-faces
					reset(view) if pend && pend.valid? && (pend.faces & faces).length>0
				end
				model.commit_operation
			end
			def curve_is_arc(curve)
				return false if !curve.is_a?Sketchup::Curve
				return true if curve.is_a?Sketchup::ArcCurve
				vs=curve.vertices.select{|v| (v.edges & curve.edges).length==2}
				angles=vs.collect{|v|
					((v.edges & curve.edges)[0].other_vertex(v).position-v.position).angle_between(((v.edges& curve.edges)[1].other_vertex(v).position-v.position))}.sort
				if angles.length>1
					return ((angles[-1]-angles[0]).abs<=0.01 )
				elsif angles.length==1 && vs.length==1
					return (((vs[0].edges & curve.edges)[0].length-(vs[0].edges & curve.edges)[1].length).abs<=0.0001)
				end
				return false
			end
			def get_info(c)
				vertices=c.vertices
				#vertices=get_vertices vertices
				reverse=false
				if c.is_a?Sketchup::ArcCurve
					vertices=get_vertices_curve vertices
					center=c.center
					if vertices.length>2
						a=(vertices[0].position-c.center).normalize.angle_between((vertices[1].position-c.center).normalize)
						a1=(vertices[2].position-center).normalize.angle_between((vertices[1].position-center).normalize)
						a=a1 if a1>a
					else
						a=(vertices[0].position-c.center).normalize.angle_between((vertices[1].position-c.center).normalize)
					end
					return [c.center, c.normal, c.radius,a,vertices]
				end
				p1=vertices[0].position
				p2=vertices[2].position
				p=vertices[1].position
				zaxis=((p2-p).normalize).cross((p1-p).normalize)
				if (p1.distance(p)-p2.distance(p)).abs>0.0001 || zaxis.length<0.00001
					vertices.reverse!
					reverse=true
					p1=vertices[0].position
					p2=vertices[2].position
					p=vertices[1].position
					zaxis=((p2-p).normalize).cross((p1-p).normalize)
					if (p1.distance(p)-p2.distance(p)).abs>0.0001 || zaxis.length<0.00001
						vertices=vertices[1..-1]
						vertices<<vertices[0]
						p1=vertices[0].position
						p2=vertices[2].position
						p=vertices[1].position
						zaxis=((p2-p).normalize).cross((p1-p).normalize)

					end
				end
				point=Geom::Point3d.new(((p1.x+p2.x)/ 2.0),((p1.y+p2.y)/ 2.0),((p1.z+p2.z)/ 2.0))
				vec= (point-p)
				line=[p,vec]
				v1 =( p-p1)
				vec1 = (v1.cross zaxis).normalize
				point1=Geom::Point3d.new(((p1.x+p.x)/ 2.0),((p1.y+p.y)/ 2.0),((p1.z+p.z)/ 2.0))
				line1=[point1,vec1]
				center=Geom::intersect_line_line(line,line1)
				r=p1.distance center
				a=(vertices[0].position-center).normalize.angle_between((vertices[1].position-center).normalize)
				a1=(vertices[2].position-center).normalize.angle_between((vertices[1].position-center).normalize)
				a=a1 if a1>a
				return [center, zaxis, r,a,vertices,reverse]
			end
			def get_vertices_curve(vertices)
			vers=[]
			#sel=Sketchup.active_model.selection.to_a
			vers<<vertices[0]
			(1..vertices.length-2).each{|i|
				next if !vertices[i-1] || !vertices[i+1] ||  !vertices[i]
				v1=vertices[i-1].position-vertices[i].position
				v2=vertices[i+1].position-vertices[i].position
				a=v1.angle_between(v2)
				vers<< vertices[i] if a.abs >0.00001 && (a-Math::PI).abs>0.00001
			}
			vers<<vertices[-1]
			return vers
		end
		end

		def self.main(mode)
			if !@line_style
				@line_style=0
				@line_style = mode if mode.between?(0,6)
			end
			@lock_z=0 if !@lock_z
			Sketchup.active_model.select_tool LineTool.new(@line_style)
		end
	end
end